#!/usr/bin/env ruby

##
# UPnPwn v0.8.0-bh
# Copyright 2008 Jonathan Squire, CISSP
##


##
## http://tools.ietf.org/html/draft-goland-http-udp-00
## http://tools.ietf.org/html/draft-cai-ssdp-v1-03
##

##
## Tested on Ubuntu 8.04LTS
##

## System Requires
require 'socket'
require 'ipaddr'

## UPnPwn requires
require 'upnp'
require 'util'

class SSDP

	attr_accessor :SSDP_port, :SSDP_address
	attr_accessor :ssdp_socket
	attr_accessor :host

  # Create the object
  def initialize(ssdp_address='239.255.255.250',ssdp_port='1900')
  
	@SSDP_address=ssdp_address
	@SSDP_port=ssdp_port

	#HEADER VARIABLES	
	@request=''
	@man=''
	#Service Type
	@st=''
	# maximum number of seconds wait before sending response
	@mx='1'

	#Notification Type
	@nt=''
	#Notification Sub-Type
	@nts=''
	#AbsoluteURI
	@usn=''
	@al=''
	#cache for T (seconds)
	@cache='60'
	
	#Setup Network Connections
	@ssdp_socket = UDPSocket.open

	# Set Send Socket TTL to 1 (why?)
	#@ssdp_socket.setsockopt(Socket::IPPROTO_IP, Socket::IP_TTL, [1].pack('i'))
	
	#Hash that will be used to hold hosts as they are found
	@host = Hash.new
  end

  def hosts
	array=Array.new
			
	@host.each {
		|key,value|
		array << key
	}
		
	return array
  end
  
  # Discover existing UPnP devices defaults to 'ssdp:all'
  def discover(service_type='ssdp:all')
	puts "Discover\n"
	
	@ssdp_socket.send(discover_msg(service_type) ,0, @SSDP_address, @SSDP_port)
	
  end

  def discover_msg(service_type='ssdp:all')
	@man='ssdp:discover'
	@st=service_type
  
	h=''
	h=h + "M-SEARCH * HTTP/1.1\r\n"
	h=h + "ST: #{@st}\r\n"
	h=h + "MX: #{@mx}\r\n"
	#Intel SDK sends quotes around MAN for some reason
	h=h + "MAN: \"#{@man}\"\r\n"
	#Spec Says no quotes
	#h=h + "MAN: #{@man}\r\n"
	h=h + "HOST: #{@SSDP_address}:#{@SSDP_port}\r\n"
	h=h + "\r\n"

	return h
  end
  

  def notify()
	@nts='ssdp:alive'
	
	@ssdp_socket.send(notify_msg(@nts) ,0, @SSDP_address, @SSDP_port)

  end
  
  def byebye
	@nts='ssdp:byebye'
	
	@ssdp_socket.send(notify_msg(@nts) ,0, @SSDP_address, @SSDP_port)
  end


  
  def notify_msg(nts)
    h=''
	h=h + "NOTIFY * HTTP/1.1\r\n"
	h=h + "HOST: #{@SSDP_address}:#{@SSDP_port}\r\n"
	h=h + "NTS: #{nts}\r\n"
	h=h + "NT: #{@nt}\r\n"
	h=h + "USN: #{@usn}\r\n"
	h=h + "AL: #{@al}\r\n"
	h=h + "Cache-Control: max-age = #{@cache}\r\n"
	h=h + "\r\n"

	return h
  end
  
  ## Main loop for handling incoming packets on a socket
  ## by default it will use the instance socket
  def server_loop(_socket=@ssdp_socket)
        puts "Entering server_loop"

        @host = Hash.new

        loop do
                msg, info = _socket.recvfrom(1024)

                ip=info[3]
                src_port=info[1]

                if (@host[ip].nil?)
                  @host[ip]=UPnPHost.new
                  @host[ip].msg_count=0
                end

                @host[ip].ip=ip
                @host[ip].msg_count=@host[ip].msg_count+1
                @host[ip].name=info[2]
                @host[ip].save_msg(msg)

                ## Show Status
                #Clear the screen
                clear_screen
                puts "=======SERVER_LOOP============"
				puts "Hosts (#{hosts.length}): #{hosts * ", "}"
                print_hosts(@host)
                puts "\n\n"

                puts "=== New Message from #{@host[ip].ip}:#{info[1]} Size: #{msg.size} ==="
                @host[ip].print
                puts "=== END ==="
        end
  end
  
  def process_message(_socket=@ssdp_socket)
    _msg, _info = _socket.recvfrom(1024)
	_ip=_info[3]
	_src_port=_info[1]
	
	#If this is a new host, create a new Hash Key and add UPnPHost to it
	if (@host[_ip].nil?)
		@host[_ip]=UPnPHost.new
		@host[_ip].msg_count=0
	end

	@host[_ip].ip=_ip
	@host[_ip].msg_count=@host[_ip].msg_count+1
	@host[_ip].name=_info[2]
	@host[_ip].save_msg(_msg)
	
  end

  def ssdp_listen
	# This HASH will contain one entry for each new IP that is found in the form:
	# hosts["x.x.x.x"]
	# the hash key will point to a UPnPHost object that represents the host
	@host = Hash.new
  
	@listen_address =  IPAddr.new(@SSDP_address).hton + IPAddr.new("0.0.0.0").hton

	sock = UDPSocket.new
	sock.bind(Socket::INADDR_ANY, @SSDP_port)
	
	sock.setsockopt(Socket::IPPROTO_IP, Socket::IP_ADD_MEMBERSHIP, @listen_address)
	
	#Disable reverse lookups
	Socket.do_not_reverse_lookup = true

	#Do a quick discovery for IGDs
	sock.send(discover_msg("urn:schemas-upnp-org:device:InternetGatewayDevice:1") ,0, @SSDP_address, @SSDP_port)
	
	server_loop(sock)
	
  end

 
 #
 # Refactor
 # returns a url pointing to the first IGD found
 # 
 # This is basically the same as a normal server loop, except it doesn't join a multicast group, and exists when it finds a location URL
   def find_IGD
	# This HASH will contain one entry for each new IP that is found in the form:
	# hosts["x.x.x.x"]
	# the hash key will point to a UPnPHost object that represents the host
	@host = Hash.new
  
	sock = UDPSocket.new
	sock.bind(Socket::INADDR_ANY, @SSDP_port)
		
	#Disable reverse lookups
	Socket.do_not_reverse_lookup = true

	#Do a quick discovery for IGDs
	sock.send(discover_msg("urn:schemas-upnp-org:device:InternetGatewayDevice:1") ,0, @SSDP_address, @SSDP_port)
	
    @host = Hash.new

	_socket=sock
    loop do
		msg, info = _socket.recvfrom(1024)

		ip=info[3]
		src_port=info[1]
		
		if (@host[ip].nil?)
			@host[ip]=UPnPHost.new
			@host[ip].msg_count=0
		end

		@host[ip].ip=ip
		@host[ip].msg_count=@host[ip].msg_count+1
		@host[ip].name=info[2]
		@host[ip].save_msg(msg)
	
		#if we find an IGD we are done
		if (!@host[ip].location.nil?)
			return @host[ip].location
		end
    end	
  end

  
  def print_hosts(hosts=@hosts)
	hosts.each do |key,value|
	puts "#{value.short_info}"
	#value.print
	end
	
  end
  
end
  

if __FILE__ == $0
  puts "Testing " + $0 + "\n" 
end
